/*
 * Copyright (C) 2020 Tobias Brunner
 *
 * Copyright (C) secunet Security Networks AG
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

/**
 * Implementation of the SHAKE128/256 XOF algorithm using OpenSSL.
 *
 * @defgroup openssl_xof openssl_xof
 * @{ @ingroup openssl_p
 */

#ifndef OPENSSL_XOF_H_
#define OPENSSL_XOF_H_

#include <library.h>

/**
 * Creates a new xof_t object.
 *
 * @param algorithm		XOF algorithm to create
 * @return				object, NULL if not supported
 */
xof_t *openssl_xof_create(ext_out_function_t algorithm);

#endif /** OPENSSL_XOF_H_ @}*/
