/* Object identifiers (OIDs) used by strongSwan
 * Copyright (C) 2003-2008 Andreas Steffen
 * 
 * This file has been automatically generated by the script oid.pl
 * Do not edit manually!
 */

#include <utils/utils.h>

#ifndef OID_H_
#define OID_H_

typedef struct {
    u_char octet;
    u_int  next;
    u_int  down;
    u_int  level;
    const u_char *name;
} oid_t;

extern const oid_t oid_names[];

#define OID_UNKNOWN							-1
#define OID_NAME_DISTINGUISHER				6
#define OID_PILOT_USERID					16
#define OID_PILOT_DOMAIN_COMPONENT			17
#define OID_COMMON_NAME						20
#define OID_SURNAME							21
#define OID_SERIAL_NUMBER					22
#define OID_COUNTRY							23
#define OID_LOCALITY						24
#define OID_STATE_OR_PROVINCE				25
#define OID_STREET_ADDRESS					26
#define OID_ORGANIZATION					27
#define OID_ORGANIZATION_UNIT				28
#define OID_TITLE							29
#define OID_DESCRIPTION						30
#define OID_POSTAL_ADDRESS					31
#define OID_POSTAL_CODE						32
#define OID_USER_CERTIFICATE				33
#define OID_NAME							34
#define OID_GIVEN_NAME						35
#define OID_INITIALS						36
#define OID_UNIQUE_IDENTIFIER				37
#define OID_DN_QUALIFIER					38
#define OID_DMD_NAME						39
#define OID_PSEUDONYM						40
#define OID_ROLE							41
#define OID_ORGANIZATION_ID					42
#define OID_SUBJECT_KEY_ID					45
#define OID_KEY_USAGE						46
#define OID_SUBJECT_ALT_NAME				48
#define OID_BASIC_CONSTRAINTS				50
#define OID_CRL_NUMBER						51
#define OID_CRL_REASON_CODE					52
#define OID_DELTA_CRL_INDICATOR				55
#define OID_ISSUING_DIST_POINT				56
#define OID_NAME_CONSTRAINTS				58
#define OID_CRL_DISTRIBUTION_POINTS			59
#define OID_CERTIFICATE_POLICIES			60
#define OID_ANY_POLICY						61
#define OID_POLICY_MAPPINGS					62
#define OID_AUTHORITY_KEY_ID				63
#define OID_POLICY_CONSTRAINTS				64
#define OID_EXTENDED_KEY_USAGE				65
#define OID_FRESHEST_CRL					67
#define OID_INHIBIT_ANY_POLICY				68
#define OID_TARGET_INFORMATION				69
#define OID_NO_REV_AVAIL					70
#define OID_CAMELLIA128_CBC					81
#define OID_CAMELLIA192_CBC					82
#define OID_CAMELLIA256_CBC					83
#define OID_RSA_ENCRYPTION					96
#define OID_MD5_WITH_RSA					98
#define OID_SHA1_WITH_RSA					99
#define OID_RSAES_OAEP						100
#define OID_MGF1							101
#define OID_RSASSA_PSS						103
#define OID_SHA256_WITH_RSA					104
#define OID_SHA384_WITH_RSA					105
#define OID_SHA512_WITH_RSA					106
#define OID_SHA224_WITH_RSA					107
#define OID_PBE_MD5_DES_CBC					109
#define OID_PBE_SHA1_DES_CBC				110
#define OID_PBKDF2							111
#define OID_PBES2							112
#define OID_PKCS7_DATA						114
#define OID_PKCS7_SIGNED_DATA				115
#define OID_PKCS7_ENVELOPED_DATA			116
#define OID_PKCS7_SIGNED_ENVELOPED_DATA		117
#define OID_PKCS7_DIGESTED_DATA				118
#define OID_PKCS7_ENCRYPTED_DATA			119
#define OID_EMAIL_ADDRESS					121
#define OID_UNSTRUCTURED_NAME				122
#define OID_PKCS9_CONTENT_TYPE				123
#define OID_PKCS9_MESSAGE_DIGEST			124
#define OID_PKCS9_SIGNING_TIME				125
#define OID_CHALLENGE_PASSWORD				127
#define OID_UNSTRUCTURED_ADDRESS			128
#define OID_EXTENSION_REQUEST				129
#define OID_X509_CERTIFICATE				132
#define OID_PBE_SHA1_RC4_128				136
#define OID_PBE_SHA1_RC4_40					137
#define OID_PBE_SHA1_3DES_CBC				138
#define OID_PBE_SHA1_3DES_2KEY_CBC			139
#define OID_PBE_SHA1_RC2_CBC_128			140
#define OID_PBE_SHA1_RC2_CBC_40				141
#define OID_P12_KEY_BAG						144
#define OID_P12_PKCS8_KEY_BAG				145
#define OID_P12_CERT_BAG					146
#define OID_P12_CRL_BAG						147
#define OID_MD5								152
#define OID_HMAC_SHA1						153
#define OID_HMAC_SHA224						154
#define OID_HMAC_SHA256						155
#define OID_HMAC_SHA384						156
#define OID_HMAC_SHA512						157
#define OID_HMAC_SHA512_224					158
#define OID_HMAC_SHA512_256					159
#define OID_3DES_EDE_CBC					161
#define OID_EC_PUBLICKEY					165
#define OID_C2PNB163V1						168
#define OID_C2PNB163V2						169
#define OID_C2PNB163V3						170
#define OID_C2PNB176W1						171
#define OID_C2PNB191V1						172
#define OID_C2PNB191V2						173
#define OID_C2PNB191V3						174
#define OID_C2PNB191V4						175
#define OID_C2PNB191V5						176
#define OID_C2PNB208W1						177
#define OID_C2PNB239V1						178
#define OID_C2PNB239V2						179
#define OID_C2PNB239V3						180
#define OID_C2PNB239V4						181
#define OID_C2PNB239V5						182
#define OID_C2PNB272W1						183
#define OID_C2PNB304W1						184
#define OID_C2PNB359V1						185
#define OID_C2PNB368W1						186
#define OID_C2PNB431R1						187
#define OID_PRIME192V1						189
#define OID_PRIME192V2						190
#define OID_PRIME192V3						191
#define OID_PRIME239V1						192
#define OID_PRIME239V2						193
#define OID_PRIME239V3						194
#define OID_PRIME256V1						195
#define OID_ECDSA_WITH_SHA1					197
#define OID_ECDSA_WITH_SHA224				199
#define OID_ECDSA_WITH_SHA256				200
#define OID_ECDSA_WITH_SHA384				201
#define OID_ECDSA_WITH_SHA512				202
#define OID_MS_CERT_TYPE_EXT				215
#define OID_MS_SMARTCARD_LOGON				216
#define OID_USER_PRINCIPAL_NAME				217
#define OID_STRONGSWAN						223
#define OID_TCGID							231
#define OID_BLOWFISH_CBC					235
#define OID_AUTHORITY_INFO_ACCESS			279
#define OID_IP_ADDR_BLOCKS					281
#define OID_POLICY_QUALIFIER_CPS			284
#define OID_POLICY_QUALIFIER_UNOTICE		285
#define OID_SERVER_AUTH						287
#define OID_CLIENT_AUTH						288
#define OID_OCSP_SIGNING					295
#define OID_XMPP_ADDR						301
#define OID_AUTHENTICATION_INFO				305
#define OID_ACCESS_IDENTITY					306
#define OID_CHARGING_IDENTITY				307
#define OID_GROUP							308
#define OID_OCSP							311
#define OID_BASIC							312
#define OID_NONCE							313
#define OID_CRL								314
#define OID_RESPONSE						315
#define OID_NO_CHECK						316
#define OID_ARCHIVE_CUTOFF					317
#define OID_SERVICE_LOCATOR					318
#define OID_CA_ISSUERS						319
#define OID_IKE_INTERMEDIATE				324
#define OID_DES_CBC							328
#define OID_SHA1							329
#define OID_SHA1_WITH_RSA_OIW				330
#define OID_ECGDSA_PUBKEY					349
#define OID_ECGDSA_SIG_WITH_RIPEMD160		352
#define OID_ECGDSA_SIG_WITH_SHA1			353
#define OID_ECGDSA_SIG_WITH_SHA224			354
#define OID_ECGDSA_SIG_WITH_SHA256			355
#define OID_ECGDSA_SIG_WITH_SHA384			356
#define OID_ECGDSA_SIG_WITH_SHA512			357
#define OID_ED25519							378
#define OID_ED448							379
#define OID_SECT163K1						383
#define OID_SECT163R1						384
#define OID_SECT239K1						385
#define OID_SECT113R1						386
#define OID_SECT113R2						387
#define OID_SECP112R1						388
#define OID_SECP112R2						389
#define OID_SECP160R1						390
#define OID_SECP160K1						391
#define OID_SECP256K1						392
#define OID_SECT163R2						393
#define OID_SECT283K1						394
#define OID_SECT283R1						395
#define OID_SECT131R1						396
#define OID_SECT131R2						397
#define OID_SECT193R1						398
#define OID_SECT193R2						399
#define OID_SECT233K1						400
#define OID_SECT233R1						401
#define OID_SECP128R1						402
#define OID_SECP128R2						403
#define OID_SECP160R2						404
#define OID_SECP192K1						405
#define OID_SECP224K1						406
#define OID_SECP224R1						407
#define OID_SECP384R1						408
#define OID_SECP521R1						409
#define OID_SECT409K1						410
#define OID_SECT409R1						411
#define OID_SECT571K1						412
#define OID_SECT571R1						413
#define OID_AES128_CBC						422
#define OID_AES128_GCM						423
#define OID_AES128_CCM						424
#define OID_AES192_CBC						425
#define OID_AES192_GCM						426
#define OID_AES192_CCM						427
#define OID_AES256_CBC						428
#define OID_AES256_GCM						429
#define OID_AES256_CCM						430
#define OID_SHA256							432
#define OID_SHA384							433
#define OID_SHA512							434
#define OID_SHA224							435
#define OID_SHA3_224						438
#define OID_SHA3_256						439
#define OID_SHA3_384						440
#define OID_SHA3_512						441
#define OID_ECDSA_WITH_SHA3_224				445
#define OID_ECDSA_WITH_SHA3_256				446
#define OID_ECDSA_WITH_SHA3_384				447
#define OID_ECDSA_WITH_SHA3_512				448
#define OID_RSASSA_PKCS1V15_WITH_SHA3_224	449
#define OID_RSASSA_PKCS1V15_WITH_SHA3_256	450
#define OID_RSASSA_PKCS1V15_WITH_SHA3_384	451
#define OID_RSASSA_PKCS1V15_WITH_SHA3_512	452
#define OID_NS_REVOCATION_URL				458
#define OID_NS_CA_REVOCATION_URL			459
#define OID_NS_CA_POLICY_URL				460
#define OID_NS_COMMENT						461
#define OID_EMPLOYEE_NUMBER					464
#define OID_PKI_MESSAGE_TYPE				470
#define OID_PKI_STATUS						471
#define OID_PKI_FAIL_INFO					472
#define OID_PKI_SENDER_NONCE				473
#define OID_PKI_RECIPIENT_NONCE				474
#define OID_PKI_TRANS_ID					475
#define OID_TPM_MANUFACTURER				481
#define OID_TPM_MODEL						482
#define OID_TPM_VERSION						483
#define OID_TPM_ID_LABEL					484

#define OID_MAX								485

#endif /* OID_H_ */
