/*
 * Copyright (C) 2015 Martin Willi
 *
 * Copyright (C) secunet Security Networks AG
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

/**
 * @defgroup chapoly_drv_portable chapoly_drv_portable
 * @{ @ingroup chapoly
 */

#include "chapoly_drv.h"

#ifndef CHAPOLY_DRV_PORTABLE_H_
#define CHAPOLY_DRV_PORTABLE_H_

/**
 * Create a chapoly_drv_portable instance.
 */
chapoly_drv_t *chapoly_drv_portable_create();

#endif /** CHAPOLY_drv_PORTABLE_H_ @}*/
